/******************************************************************************************************/
/********************  Alorium XLR8, EasyVR 3.0, Adafruit Bluefruit SPI Integration *******************/
/******************************************************************************************************/
/*  This program allows users to create a voice-activate keyboard that can send desired keypresses and
    keycodes.

                           Alorium XLR8 Board  |  http://aloriumtech.com
                            Easy VR 3.0 Board  |  http://Veear.eu
                         Bluefruit SPI Shield  |  http://adafruit.com

  ARDUINO IDE INSTRUCTIONS
    In Arduino IDE, add "Adafruit Bluefruit NRF41" and "EasyVR" libraries.  Boards tested include XLR8
    and Sparkfun Redboard

  EASYVR V3.0 or SHIELD SETUP INSTRUCTIONS
    If using the shield note jumper location (should be in SW position.  Hot air rework component marked
    "101" (100 Ω resistor) connecting TX/RX pins 12/13 to 8/9 to avoid conflict with Bluefruit Shield.
    Alternatively, remove mode jumper and wirewrap GND, 5V to arduino Gnd and 5V, and module Tx/Rx to
    Arduino pins 8/9 (Tx -> 8, Rx -> 9)

  BLUEFRUIT SHIELD SPI SETUP INSTRUCTIONS
    Solder 2x6, 2x8, 2x8, and 2x10 pass-through headers on the shield.  On the underside of the sheild,
    break the trace between CS and pin 8 and reattach CS to pin 10 with wirewrap or other method.

    Program by Mark Hughes for AllAboutCircuits.com.
    Code is based and heavily influenced by examples provided by Adafruit.com (lady ada, Kevin Townsend)
    and Veear.eu (library examples)

  Additional Information At End of File
*/
// Software Includes
#include <SPI.h>                              // #include &lt;SPI.h&gt;
#include <SoftwareSerial.h>                   // #include &lt;SoftwareSerial.h&gt;
#include <EasyVR.h>                           // #include &lt;EasyVR.h&gt;
#include <Adafruit_ATParser.h>                // #include &lt;Adafruit_ATParser.h&gt;
#include <Adafruit_BLE.h>                     // #include &lt;Adafruit_BLE.h&gt;
#include <Adafruit_BLEGatt.h>                 // #include &lt;Adafruit_BLEGatt.h&gt;
#include <Adafruit_BluefruitLE_SPI.h>         // #include &lt;Adafruit_BluefruitLE_SPI.h&gt;

// Pin Definitions
#define BLUEFRUIT_SPI_SCK     13                // Attach 100Ω Pull-Down Resistor
#define BLUEFRUIT_SPI_MISO    12                // 
#define BLUEFRUIT_SPI_MOSI    11                // 
#define BLUEFRUIT_SPI_CS      10                // Default pin 8 conflicts with EasyVR shield
#define BLUEFRUIT_SPI_IRQ     7
#define BLUEFRUIT_SPI_RST     4
#define EasyVR_RX             8                 // Must move resistors on shield from pins 12/13
#define EasyVR_TX             9                 // to pins 8/9 to avoid SPI conflict

// Constants
#define BUFSIZE                        128      // Size of the read buffer for incoming data
#define VERBOSE_MODE                   false    // If set to 'true' enables bluefruit debug output
#define pcSerial SERIAL_PORT_MONITOR            // Used to debug and send keystrokes
#define FACTORYRESET_ENABLE 0                   // Set to 1 to reset bluefruit
#define MINIMUM_FIRMWARE_VERSION    "0.6.6"

// Global Variables
volatile char directions;                       // Shortcut to pass value without rewriting function
volatile char repetitions;                      // Shortcut to pass value without rewriting function

// Create Easy VR on a Software Serial Port
SoftwareSerial port(EasyVR_RX, EasyVR_TX);      // Create a software serial port
EasyVR easyvr(port);                            // Attach EasyVR to software serial port

//Create Bluefruit object with SPI
Adafruit_BluefruitLE_SPI ble(BLUEFRUIT_SPI_SCK, BLUEFRUIT_SPI_MISO, BLUEFRUIT_SPI_MOSI,
                             BLUEFRUIT_SPI_CS, BLUEFRUIT_SPI_IRQ, BLUEFRUIT_SPI_RST);

void error(const __FlashStringHelper*errorText) { // Print errors using pointer
  Serial.println();
  Serial.println(errorText);

}


//Groups and Commands
enum Groups { GROUP_0  = 0,};                 // Default start keyword "Robot"
enum Group0 { G0_COMPUTER = 0,};              // Trained keyword "Computer"

//Grammars and Words
enum Wordsets {SET_1 = -1, SET_2 = -2, SET_3 = -3}; // Action, Direction, Scale
enum Wordset1 { S1_ACTION = 0, S1_MOVE = 1, S1_TURN = 2, S1_RUN = 3, S1_LOOK = 4, S1_ATTACK = 5, S1_STOP = 6, S1_HELLO = 7 };
enum Wordset2 { S2_LEFT = 0, S2_RIGHT = 1, S2_UP = 2, S2_DOWN = 3, S2_FORWARD = 4, S2_BACKWARD = 5 };
enum Wordset3 { S3_ZERO = 0, S3_ONE = 1, S3_TWO = 2, S3_THREE = 3, S3_FOUR = 4, S3_FIVE = 5, S3_SIX = 6, S3_SEVEN = 7, S3_EIGHT = 8, S3_NINE = 9, S3_TEN = 10};

int8_t group, idx;                            // use negative group for wordsets

void setup()
{
  pcSerial.begin(115200);                     // setup PC serial port

  Serial.print(F("Initialising the Bluefruit LE module: "));
  if ( !ble.begin(VERBOSE_MODE) )
  {
    error(F("Couldn't find Bluefruit, make sure it's in CoMmanD mode & check wiring?"));
  }
  Serial.println( F("OK!") );

  if ( FACTORYRESET_ENABLE )
  {
    /* Perform a factory reset to make sure everything is in a known state */
    Serial.println(F("Performing a factory reset: "));
    if ( ! ble.factoryReset() ) {
      error(F("Couldn't factory reset"));
    } // end ! ble.factoryReset()
  } // end FACTORYRESET_ENABLE
  ble.echo(false);
  Serial.println("Requesting Bluefruit info:");
  ble.info();   // Print Bluefruit information

  Serial.println(F("Setting device name to 'AACVoiceKeyboard': "));
  if (! ble.sendCommandCheckOK(F( "AT+GAPDEVNAME=AACVoiceKeyboard" )) ) {
    pcSerial.println(F("Could not set device name!"));
  }

  /* Enable HID Service */
  Serial.println(F("Enable HID Service (including Keyboard): "));
  if ( ble.isVersionAtLeast(MINIMUM_FIRMWARE_VERSION) )
  { if ( !ble.sendCommandCheckOK(F( "AT+BleHIDEn=On" ))) {
      pcSerial.println(F("Could not enable Keyboard"));
    }
    else
    { if (! ble.sendCommandCheckOK(F( "AT+BleKeyboardEn=On" ))) {
        pcSerial.println(F("Could not enable Keyboard")); // End if
      }
    }
  } //end if(min version)

  /* Add or remove service requires a reset */
  Serial.println(F("Performing a SW reset (service changes require a reset): "));
  if (! ble.reset() ) {
    pcSerial.println(F("Couldn't reset"));
  }

  Serial.println();
  Serial.println(F("Go to your devices Bluetooth settings to pair your device"));
  Serial.println(F("then open an application that accepts keyboard input"));
  Serial.println();

bridge:
  // bridge mode?
  int mode = easyvr.bridgeRequested(pcSerial);
  switch (mode)
  {
    case EasyVR::BRIDGE_NONE:
      /* attempt to communicate at 115200 baud.  If not, establish communication at 9600 baud and change settings
      */
      port.begin(115200);                                         // Communicate at 115200 baud
      if (!easyvr.detect()) {                                     // If easyVR doesn't respond,
        port.end();                                               // close port
        port.begin(9600);                                         // Communicate at 9600 baud
        if (!easyvr.detect()) {                                   // If easy VR doesn't respond,error
          error(F("No EasyVR Detected at 9600 or 115200 Baud. Check Jumper Settings."));
        } else {                                                  // If easy VR does respond, change baud.
          if (easyvr.changeBaudrate(EasyVR::B115200)) {
            pcSerial.println("Baud rate increased from 9600->115200 Baud");
            port.end();                                           // Need to stop 9600 baud communication
            port.begin(115200);                                   // Reestablished at 115200 baud until next reboot.
          } // end if(easyvr.changebaurrate)
        } // end else
      } // end if(!easyvr.detect()

      pcSerial.println(F("Bridge not requested, run normally"));
      pcSerial.println(F("---"));
      break;

    case EasyVR::BRIDGE_NORMAL:
      port.begin(9600);           // setup EasyVR serial port (low speed)
      // soft-connect the two serial ports (PC and EasyVR)
      easyvr.bridgeLoop(pcSerial);
      // resume normally if aborted
      pcSerial.println(F("Bridge connection aborted"));
      pcSerial.println(F("----------------"));
      break;

    case EasyVR::BRIDGE_BOOT:
      // setup EasyVR serial port (high speed)
      port.begin(115200);
      pcSerial.begin(115200);
      // soft-connect the two serial ports (PC and EasyVR)
      easyvr.bridgeLoop(pcSerial);
      // resume normally if aborted
      pcSerial.println(F("Bridge connection aborted"));
      pcSerial.println(F("---"));
      break;
  } // end Switch mode

  // initialize EasyVR
  while (!easyvr.detect())
  {
    pcSerial.println(F("EasyVR not detected!"));
    for (int i = 0; i < 10; ++i)
    {
      if (pcSerial.read() == '?')
        goto bridge;
      delay(100);
    }
  }

  pcSerial.print(F("EasyVR detected, version "));
  pcSerial.println(easyvr.getID());

  if (easyvr.getID() < EasyVR::EASYVR3)
    easyvr.setPinOutput(EasyVR::IO1, LOW); // Shield 2.0 LED off

  if (easyvr.getID() < EasyVR::EASYVR)
    pcSerial.print(F(" = VRbot module"));
  else if (easyvr.getID() < EasyVR::EASYVR2)
    pcSerial.print(F(" = EasyVR module"));
  else if (easyvr.getID() < EasyVR::EASYVR3)
    pcSerial.print(F(" = EasyVR 2 module"));
  else
    pcSerial.print(F(" = EasyVR 3 module"));
  pcSerial.print(F(", FW Rev."));
  pcSerial.println(easyvr.getID() & 7);

  easyvr.setDelay(0); // speed-up replies

  easyvr.setTimeout(5);
  easyvr.setLanguage(0);

  group = EasyVR::TRIGGER;                  //<-- start group (customize)
}

void loop(void)
{
  if (easyvr.getID() < EasyVR::EASYVR3)
    easyvr.setPinOutput(EasyVR::IO1, HIGH); // LED on (listening)

  if (group < 0) // SI wordset/grammar
  {
    switch (-group)
    {
      case 2: pcSerial.print("Say a Direction:  "); break;
      case 3: pcSerial.print("Say a Distance:   "); break;
    }
    easyvr.recognizeWord(-group);
  }
  else // SD group
  {
    pcSerial.println("Say a trigger word: \"Computer\" or \"Robot\" ");
    easyvr.recognizeCommand(group);
  }

  do
  {
    // allows Commander to request bridge on Zero (may interfere with user protocol)
    if (pcSerial.read() == '?')
    {
      setup(); return;
    }
  }
  while (!easyvr.hasFinished());

  if (easyvr.getID() < EasyVR::EASYVR3)
    easyvr.setPinOutput(EasyVR::IO1, LOW); // LED off

  idx = easyvr.getWord();
  if (idx == 0 && group == EasyVR::TRIGGER)         // detect built in word Robot
  {
    easyvr.playSound(0, EasyVR::VOL_FULL);          // beep

    pcSerial.println("Word: ROBOT");                // print debug message
    easyvr.playSound(3, EasyVR::VOL_FULL);          // Custom Beep (0 default)
    Serial.println(F("Enter the character(s) to send:"));
    Serial.println(F("- \\r for Enter"));
    Serial.println(F("- \\n for newline"));
    Serial.println(F("- \\t for tab"));
    Serial.println(F("- \\b for backspace"));

    Serial.print(F("keyboard > "));
    char keys[BUFSIZE + 1];
    getUserInput(keys, BUFSIZE);
    easyvr.playSound(2, EasyVR::VOL_FULL);          // Custom Beep "down up down"
    Serial.print("\nSending "); Serial.println(keys);
    ble.print("AT+BleKeyboard="); ble.println(keys);
    keys[BUFSIZE + 1] = {};

    if ( ble.waitForOK() )
    {
      Serial.println( F("OK!") );
    }
    else {
      error( F("FAILED!") );
      easyvr.playSound(1, EasyVR::VOL_FULL);
    }

    return;
  }

  else if (idx >= 0)
  {
    // beep
    easyvr.playSound(0, EasyVR::VOL_FULL);
    // print debug message
    uint8_t flags = 0, num = 0;
    char name[32] = "";
    pcSerial.print("Word: ");
    pcSerial.print(idx);
    if (easyvr.dumpGrammar(-group, flags, num))
    {
      for (uint8_t pos = 0; pos < num; ++pos)
      {
        if (!easyvr.getNextWordLabel(name))
          break;
        if (pos != idx)
          continue;
        pcSerial.print(F(" = "));
        pcSerial.println(name);
        String name[32] = "";
        break;
      } // end for uint_8t
    } // end easyvr.dumpGrammar
    action();  // get the words and numbers

    for (int n = 1; n <= repetitions; n++) {
      ble.print(F("AT+BLEKEYBOARDCODE="));
      pcSerial.print(F("AT+BLEKEYBOARDCODE="));
      switch (directions) {
        case 1: //Left Arrow
          ble.println(F("00-00-50-00-00-00-00"));
          pcSerial.println(F("00-00-50-00-00-00-00"));
          break;

        case 2: //Right Arrow
          ble.println(F("00-00-4F-00-00-00-00"));
          pcSerial.println(F("00-00-4F-00-00-00-00"));
          break;

        case 3: //Up Torch (Page Up)
          ble.println(F("00-00-4B-00-00-00-00"));
          pcSerial.println(F("00-00-4B-00-00-00-00"));
          break;

        case 4: //Down Torch (Page Down)
          ble.println(F("00-00-4E-00-00-00-00"));
          pcSerial.println(F("00-00-4E-00-00-00-00"));
          break;

        case 5: //Forward (Up Arrow)
          ble.println(F("00-00-52-00-00-00-00"));
          pcSerial.println(F("00-00-52-00-00-00-00"));
          break;

        case 6: //Backward(Down Arrow)
          ble.println(F("00-00-51-00-00-00-00"));
          pcSerial.println(F("00-00-51-00-00-00-00"));
          break;

      } // End Switch (Directions)
      delay(25);  // Need some delay between repetitive keypress
      ble.println(F("AT+BLEKEYBOARDCODE=00-00"));
      pcSerial.println(F("AT+BLEKEYBOARDCODE=00-00"));
      repetitions = 0;        // Reset Repetitions to prevent unintended loop trigger
    } // End for loop (repetitions)
    return;
  } // end else if (idx)
  idx = easyvr.getCommand();
  if (idx >= 0)
  {
    easyvr.playSound(0, EasyVR::VOL_FULL);  // beep
    // print debug message
    uint8_t train = 0;
    char name[32];
    pcSerial.print("Command: ");
    //    pcSerial.print(idx);
    if (easyvr.dumpCommand(group, idx, name, train)){
      //      pcSerial.println(name); // "THREE", "DOWN", etc...
    }
      else
        pcSerial.println();
    action();
  } // end if(idx)
  else // errors and timeout
  {
    if (easyvr.isTimeout())
      pcSerial.println(F("Timed out, try again..."));
    int16_t err = easyvr.getError();
    if (err >= 0)
    {
      pcSerial.print("Easy VR Error: ");
      pcSerial.print(err, HEX);
      easyvr.playSound(1, EasyVR::VOL_FULL);  // Play error sound
      switch (err) {
        case 0x02: error(F("Too long")); break;
        case 0x03: error(F("Too noisy")); break;
        case 0x04: error(F("Spoken too softly")); break;
        case 0x05: error(F("Spoken too loud")); break;
        case 0x06: error(F("Spoken too soon")); break;
        case 0x07: error(F("Too complex")); break;
        case 0x08: error(F("Invalid SI weights")); break;
        case 0x09: error(F("Invalid setup")); break;
        case 0x11: error(F("Recognition failure")); break;
        case 0x12: error(F("Recognition result doubtful")); break;
        case 0x13: error(F("Recognition result maybe")); break;
        case 0x14: error(F("Invalid SD/SV template")); break;
        case 0x15: error(F("Invalid SI weights")); break;
        case 0x17: error(F("Incomplete pattern duration")); break;
        case 0x21: error(F("T2si state structure is too big")); break;
        case 0x22: error(F("T2si RSC code version/ Grammer Rom don't match")); break;
        case 0x23: error(F("Out of t2si Ram!")); break;
        case 0x24: error(F("Unexpected t2si pcSerial.print occured")); break;
        case 0x25: error(F("Ran out of time")); break;
        case 0x26: error(F("T2si bad macro or grammar parameter")); break;
        case 0x29: error(F("T2si layer size out of limits")); break;
        case 0x2A: error(F("T2si net structure incompatibility")); break;
        case 0x2B: error(F("T2si initialization not complete")); break;
        case 0x2C: error(F("T2si incorrect number of layers")); break;
        case 0x2D: error(F("T2si trigger recognised Out of Vocabulary")); break;
        case 0x2F: error(F("T2si utterance too short")); break;
        case 0x31: error(F("Play -- illegal compression level")); break;
        case 0x38: error(F("Play, erase, copy - mag doesn't exist")); break;
        case 0x39: error(F("Rec, copy - mag already exists")); break;
        case 0x4A: error(F("Bad release nubmer in speech file")); break;
        case 0x4E: error(F("Bad message data or SX technology file missing")); break;
        case 0x80: error(F("None of the above (out of grammar)")); break;
        case 0x81: error(F("Invalid data (for memory check)")); break;
        case 0xC0: error(F("No room left in software stack")); break;
        case 0xCC: error(F("T2si test mode error")); break;
      } // end switch(err)
    } // end if (err>=0)
  } // end else
} // end void loop()

void action()
{
  switch (group)
  {
    case GROUP_0:
      switch (idx)
      {
        case G0_COMPUTER:
          group = -2; // goto directions (left, right, etc...) wordset.
          break;
      }
      break;
    case SET_2:
      switch (idx) {  // directions Choice
        case S2_LEFT:     directions = 1; group = -3; break;
        case S2_RIGHT:    directions = 2; group = -3; break;
        case S2_UP:       directions = 3; group = -3; break;
        case S2_DOWN:     directions = 4; group = -3; break;
        case S2_FORWARD:  directions = 5; group = -3; break;
        case S2_BACKWARD: directions = 6; group = -3; break;
      } // end switch(idx)
      break;
    case SET_3:
      switch (idx) {  // Repetition choice
        case S3_ZERO:     repetitions = 0;  group = 0; break;
        case S3_ONE:      repetitions = 1;  group = 0; break;
        case S3_TWO:      repetitions = 2;  group = 0; break;
        case S3_THREE:    repetitions = 3;  group = 0; break;
        case S3_FOUR:     repetitions = 4;  group = 0; break;
        case S3_FIVE:     repetitions = 5;  group = 0; break;
        case S3_SIX:      repetitions = 6;  group = 0; break;
        case S3_SEVEN:    repetitions = 7;  group = 0; break;
        case S3_EIGHT:    repetitions = 8;  group = 0; break;
        case S3_NINE:     repetitions = 9;  group = 0; break;
        case S3_TEN:      repetitions = 10; group = 0; break;
      } // end switch(idx)
      break;
  } // end switch(group)
} // end action()

void getUserInput(char buffer[], uint8_t maxSize)
{
  memset(buffer, 0, maxSize);
  while ( Serial.available() == 0 ) {
    delay(1);
  }

  uint8_t count = 0;

  do
  {
    count += Serial.readBytes(buffer + count, maxSize);
    delay(2);
  } while ( (count < maxSize) && !(Serial.available() == 0) );
}
/******************************************************************************************************/
/****************************** How to Send Custom Keypresses and Macros ******************************/
/******************************************************************************************************/
/*
  To send custom macros and keypress combinations, send the command AT+BLEKEYBOARDCODE= followed by
  three to seven, dash separated bytes that consist of 2-digit hex codes that correspond to HID
  Keyboard codes (07).  Immediately follow with AT+BLEKEYBOARDCODE=00-00 to indicate that all keys
  were released.
    For example
      ble.println("AT+BLEKEYBOARDCODE=02-00-04-16-06-0C-0C"); // This sends a key combination
      ble.println("AT+BLEKEYBOARDCODE=00-00"); // This mandatory command release all depressed keys
    Generically, the full command looks like this:
      ble.println("AT+BLEKEYBOARDCODE=XX-00-YY-NN-NN-NN-NN");
        Where:
          MM (byte 1) is the hex code for a modifier key
          00 (byte 2) is a reserved hex code (always 00)
          NN (byte 3) is the desired key-press
          PP (byte 4-7) are optional additional key-presses

  Keyboard code format http://www.usb.org/developers/hidpage/Hut1_12v2.pdf  (pages 53-59)

        +-----------------------------------------------------------------------------------+
        |                              HID Keyboard Code Format                             |
        +-----------------------------------------------------------------------------------+
        |                http://www.usb.org/developers/hidpage/Hut1_12v2.pdf                |
        +-----------------------------------------------------------------------------------+
        | Byte 0x01 | Byte 0x02 | Byte 0x03 | Byte 0x04 | Byte 0x05 | Byte 0x06 | Byte 0x07 |
        +-----------+-----------+-----------+-----------+-----------+-----------+-----------+
        |  Modifer  |  Reserved | Keycode 1 | Keycode 2 | Keycode 3 | Keycode 4 | Keycode 5 |
        +-----------+-----------+-----------+-----------+-----------+-----------+-----------+
        |    0xMM   |    0x00   |    0xNN   |    0xPP   |    0xPP   |    0xPP   |    0xPP   |
        +-----------+-----------+-----------+-----------+-----------+-----------+-----------+
    +------------------------------------+    +----------------------------------------------------+
    |       Byte 01 -- Hex Modifier      |    | Modifier       Reserved       Key Code      Result |
    +------------------------------------+    +----------------------------------------------------+
    |    Modifer    |   Binary  | Hex ID |    |    00      -      00      -      04    |  |    a   |
    +---------------+-----------+--------+    +----------------------------------------+--+--------+
    |      None     | B00000000 |  0x00  |    |    02      -      00      -      04    |  |    A   |
    +---------------+-----------+--------+    +----------------------------------------+--+--------+
    |  Left Control | B00000001 |  0x01  |    |    00      -      00      -      1E    |  |    1   |
    +---------------+-----------+--------+    +----------------------------------------+--+--------+
    |   Left Shift  | B00000010 |  0x02  |    |    02      -      00      -      1E    |  |    !   |
    +---------------+-----------+--------+    +----------------------------------------------------+
    |    Left Alt   | B00000100 |  0x04  |
    +---------------+-----------+--------+
    |  Left Window  | B00001000 |  0x08  |
    +---------------+-----------+--------+
    | Right Control | B00010000 |  0x10  |
    +---------------+-----------+--------+
    |  Right Shift  | B00100000 |  0x20  |
    +---------------+-----------+--------+
    |   Right Alt   | B01000000 |  0x40  |
    +---------------+-----------+--------+
    |  Right Window | B10000000 |  0x80  |
    +---------------+-----------+--------+
*/
/******************************************************************************************************/
